/*
 Copyright (c) 2020 gematik GmbH

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */
package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.AesGcmEncryptionData;
import de.gematik.ti.sgd.domain.configuration.SgdInstanceProperties;
import de.gematik.ti.sgd.domain.json.*;
import de.gematik.ti.sgd.domain.sgdCommunication.EncryptedMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.KeyDerivationRequestSgdMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.PublicKeyEcies;
import de.gematik.ti.sgd.domain.sgdCommunication.vectors.SgdVector;
import de.gematik.ti.sgd.exceptions.server.SgdServerException;
import lombok.Data;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.RandomStringUtils;
import org.apache.commons.lang3.tuple.Pair;

import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static de.gematik.ti.sgd.components.SgdClientUtils.extractCertificatesFromListOfFilenames;
import static de.gematik.ti.sgd.components.SgdClientUtils.extractPublicKeysFromListOfCertificateFilenames;
import static de.gematik.ti.sgd.components.SgdServerUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail;

@Data
public class SchluesselGenerierungsDienst implements SgdVector.DerivationKeyRepository {
    private final SgdCryptoUtilities sgdCryptoUtilities;

    private boolean isThisSgdOne;
    private X509Certificate sgdCertificate;
    private PrivateKey sgdPrivateKey; //(S1)
    private boolean performServerSidedHCheck = true;

    private List<PublicKey> possibleCertificateSigners; //(S2)
    private List<X509Certificate> possibleOcspSignerCertificates; // kann der TSL entnommen werden

    private Map<String, byte[]> availableDerivationKeys = new HashMap<>();
    private String currentDerivationKeyName;

    private KeyPair currentHsmEciesKeyPair; //(S4)
    private KeyPair previousHsmEciesKeyPair; //(S4 alt)
    private byte[] currentAuthenticationTokenDerivationKey; // (S5)
    private byte[] previousAuthenticationTokenDerivationKey; // (S5)

    public SchluesselGenerierungsDienst() {
        sgdCryptoUtilities = SgdCryptoUtilities.getInstance();

        performEciesKeyChange();
        initAtDerivationKey();
        addRandomDerivationKey();
    }

    public SchluesselGenerierungsDienst(String currentDerivationKeyName, byte[] currentDerivationKeyValue) {
        this();

        availableDerivationKeys.put(currentDerivationKeyName, currentDerivationKeyValue);
        this.currentDerivationKeyName = currentDerivationKeyName;
    }

    public SchluesselGenerierungsDienst(X509Certificate sgdCertificate, PrivateKey sgdPrivateKey, KeyPair currentHsmEciesKeyPair) {
        sgdCryptoUtilities = SgdCryptoUtilities.getInstance();

        this.sgdCertificate = sgdCertificate;
        this.sgdPrivateKey = sgdPrivateKey;
        this.currentHsmEciesKeyPair = currentHsmEciesKeyPair;

        this.isThisSgdOne = isCertificateForSgdOne(sgdCertificate);
        initAtDerivationKey();
        addRandomDerivationKey();
    }

    public SchluesselGenerierungsDienst(SgdInstanceProperties sgdProperties, List<String> trustedSignerCertificateFiles, Map<String, byte[]> availableDerivationKeys) {
        sgdCryptoUtilities = SgdCryptoUtilities.getInstance();
        this.performServerSidedHCheck = sgdProperties.isPerformServerSidedHCheck();
        this.sgdCertificate = sgdCryptoUtilities.loadX509CertificateFromP12File(sgdProperties.getCertificateP12(), sgdProperties.getCertificateP12Password());
        this.sgdPrivateKey = sgdCryptoUtilities.loadPrivateKeyFromP12File(sgdProperties.getCertificateP12(), sgdProperties.getCertificateP12Password(), sgdProperties.getCertificateP12Password());
        this.currentHsmEciesKeyPair = sgdCryptoUtilities.generateEccKeyPair();
        this.availableDerivationKeys.put(sgdProperties.getCurrentDerivationKeyName(), sgdCryptoUtilities.parseHexStringToByteArray(sgdProperties.getCurrentDerivationKeyValue()));
        this.currentDerivationKeyName = sgdProperties.getCurrentDerivationKeyName();
        this.availableDerivationKeys.putAll(availableDerivationKeys);
        this.isThisSgdOne = isCertificateForSgdOne(sgdCertificate);

        possibleCertificateSigners = extractPublicKeysFromListOfCertificateFilenames(trustedSignerCertificateFiles);
        possibleOcspSignerCertificates = extractCertificatesFromListOfFilenames(trustedSignerCertificateFiles);
        initAtDerivationKey();
    }

    public static String generate256BitRandomString() {
        return RandomStringUtils.random((256 / 8) * 2, '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', 'a', 'b', 'c', 'd', 'e', 'f');
    }

    private boolean isCertificateForSgdOne(X509Certificate sgdCertificate) {
        return sgdCryptoUtilities.doesCertificateContainAdmissionsExtensionOid(sgdCertificate, SgdCryptoUtilities.OID_SGD1_HSM);
    }

    private EncryptedMessageContainer buildEncryptedMessageContainerWithPayloadForRequest(String payload, SgdRequest request, Map.Entry<String, String>... additionalDebugInformation) {
        final PublicKey clientEphemeralEciesKey;
        try {
            clientEphemeralEciesKey = request.getPublicKeyEcies().getPublicKey();
        } catch (RuntimeException e) {
            throw new SgdServerException("Error while extracting the PublicKey from the Request", SgdResponseStatus.PUBLIC_KEY_CORRUPTED);
        }

        final EncryptedMessage encryptedMessage = new EncryptedMessage(clientEphemeralEciesKey, payload);

        return EncryptedMessageContainer.EncryptedMessageContainerBuilder.anEncryptedMessageContainer()
                .withStatus(SgdResponseStatus.OK)
                .withEncryptedMessage(encryptedMessage)
                .withDebug_encryptedMessageCleartext(payload)
                .withDebug_clientEphemeralKey_Public(sgdCryptoUtilities.extractPublicPointCoordinatesFromPublicKey(clientEphemeralEciesKey))
                .withDebug_senderKey_Public(sgdCryptoUtilities.extractPublicPointCoordinatesFromPublicKey(encryptedMessage.getSenderPublicKey()))
                .withDebug_data(Stream.of(additionalDebugInformation).collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue)))
                .build();
    }

    private void initAtDerivationKey() {
        currentAuthenticationTokenDerivationKey = generateAtDerivationKey();
    }

    private byte[] generateAtDerivationKey() {
        // gemSpec_SGD_epa, 4.5.2, A_18026
        byte[] result = new byte[512 / 8];
        ThreadLocalRandom.current().nextBytes(result);
        return result;
    }

    private void addRandomDerivationKey() {
        // gemSpec_Krypt, 3.15.5, A_17876
        byte[] derivationKey = new byte[512 / 8];
        ThreadLocalRandom.current().nextBytes(derivationKey);
        availableDerivationKeys.put("schluesselname", derivationKey);
        this.currentDerivationKeyName = "schluesselname";
    }

    public GetPublicKeyResponse getPublicKey(SgdRequest request) {
        performGetPublicKeyVerificationAndThrowExceptionWhenFail(request, possibleCertificateSigners, possibleOcspSignerCertificates);

        PublicKeyEcies publicKeyEcies = new PublicKeyEcies(currentHsmEciesKeyPair.getPublic());

        return GetPublicKeyResponse.builder()
                .publicKeyEcies(publicKeyEcies)
                .certificate(getSgdCertificateBase64Encoded())
                .signature(publicKeyEcies.getSignatureAsBase64EncodedString(getSgdPrivateKey()))
                .build();
    }

    private String getSgdCertificateBase64Encoded() {
        try {
            return sgdCryptoUtilities.convertBytesToB64String(sgdCertificate.getEncoded());
        } catch (CertificateEncodingException e) {
            throw new RuntimeException("Error while trying to encode SGD certificate", e);
        }
    }

    public EncryptedMessageContainer getAuthenticationToken(SgdRequest request) {
        performSignatureCheckAndThrowExceptionInCaseOfFailure(request);
        performGetPublicKeyVerificationAndThrowExceptionWhenFail(request, this.possibleCertificateSigners, possibleOcspSignerCertificates);
        checkIfOwnPublicKeyEciesIsInRequest(request);

        final String[] decryptedMessageParts = new String(request.getEncryptedMessage().decryptMessage(currentHsmEciesKeyPair, previousHsmEciesKeyPair)).split(" ");
        byte[] clientChallenge = sgdCryptoUtilities.parseHexStringToByteArray(decryptedMessageParts[1]);

        final byte[] aValue = generateAuthenticationTokenAString(request);
        String authenticationToken = generateAuthenticationTokenValue(aValue, currentAuthenticationTokenDerivationKey);
        String hValue = DigestUtils.sha256Hex(aValue);
        String messagePayload = "Response " + sgdCryptoUtilities.convertToLowerCaseHexWith64ByteLength(clientChallenge) + " " + hValue + " " + authenticationToken;

        if (this.performServerSidedHCheck) {
            if (decryptedMessageParts.length < 3) {
                throw new SgdServerException("No H-Value found in Encrypted-Message", SgdResponseStatus.MISSING_DATA);
            }
            if (!hValue.equals(decryptedMessageParts[2])) {
                throw new SgdServerException("H-Value in Encrypted-Message not correct!", SgdResponseStatus.DECRYPTION_FAIL);
            }
        }

        return buildEncryptedMessageContainerWithPayloadForRequest(messagePayload, request,
                Pair.of("hRawValueAsHexString", sgdCryptoUtilities.convertByteArrayToHexString(aValue)),
                Pair.of("authenticationToken", authenticationToken),
                Pair.of("clientChallenge", sgdCryptoUtilities.convertToLowerCaseHexWith64ByteLength(clientChallenge)),
                Pair.of("currentAuthenticationTokenDerivationKey", sgdCryptoUtilities.convertByteArrayToHexString(currentAuthenticationTokenDerivationKey)));
    }

    private String generateAuthenticationTokenValue(byte[] aValue, byte[] authenticationTokenDerivationKey) {
        byte[] authenticationTokenRaw = sgdCryptoUtilities.performHkdf(authenticationTokenDerivationKey, aValue, 256);
        return "AT" + sgdCryptoUtilities.convertToLowerCaseHexWith64ByteLength(authenticationTokenRaw);
    }

    private byte[] generateAuthenticationTokenAString(SgdRequest request) {
        try {
            return ArrayUtils.addAll(
                    request.getPublicKeyEcies().getBytes(),
                    request.getCertificateAsX509().getEncoded());
        } catch (CertificateEncodingException e) {
            throw new RuntimeException(e);
        }
    }

    public boolean isThisSgdOne() {
        return isThisSgdOne;
    }

    private void checkIfOwnPublicKeyEciesIsInRequest(SgdRequest request) {
        String currentHash = DigestUtils.sha256Hex(new PublicKeyEcies(currentHsmEciesKeyPair.getPublic()).getEciesPublicKeyAsString());

        String receivedHash = isThisSgdOne ?
                request.getPublicKeyEcies().getSgd1PublicKeyHash()
                : request.getPublicKeyEcies().getSgd2PublicKeyHash();

        if (currentHash.equals(receivedHash)) {
            // hash is valid
            return;
        }
        if (previousHsmEciesKeyPair == null) {
            throw new SgdServerException("Received wrong ECIES-Public-Key-Hash!", SgdResponseStatus.PUBLIC_KEY_CORRUPTED);
        }

        String previousHash = DigestUtils.sha256Hex(new PublicKeyEcies(previousHsmEciesKeyPair.getPublic()).getEciesPublicKeyAsString());
        if (previousHash.equals(receivedHash)) {
            // hash is valid
            return;
        } else {
            throw new SgdServerException("Received wrong ECIES-Public-Key-Hash!", SgdResponseStatus.PUBLIC_KEY_CORRUPTED);
        }
    }

    private void performAtVerification(SgdRequest sgdRequest, String givenAuthenticationToken) {
        final byte[] aValue = generateAuthenticationTokenAString(sgdRequest);
        String currentGeneratedAuthenticationToken = generateAuthenticationTokenValue(aValue, currentAuthenticationTokenDerivationKey);
        if (currentGeneratedAuthenticationToken.equals(givenAuthenticationToken)) {
            return;
        }

        if (previousAuthenticationTokenDerivationKey != null) {
            String previousGeneratedAuthenticationToken = generateAuthenticationTokenValue(aValue, previousAuthenticationTokenDerivationKey);
            if (previousGeneratedAuthenticationToken.equals(givenAuthenticationToken)) {
                return;
            }
        }

        throw new SgdServerException("Unable to verify given Authentication-Token", SgdResponseStatus.KEY_DERIVATION_FAIL);
    }

    private void performSignatureCheckAndThrowExceptionInCaseOfFailure(SgdRequest request) {
        byte[] payload = request.getPublicKeyEcies().getBytes();
        PublicKey publicKey = request.getCertificateAsX509().getPublicKey();
        byte[] signatureInQuestion = sgdCryptoUtilities.convertB64StringToBytes(request.getSignature());

        try {
            sgdCryptoUtilities.verifySignatureAndThrowExceptionWhenFail(payload, publicKey, signatureInQuestion);
        } catch (Exception e) {
            throw new SgdServerException("Error while verifying ECIES-Signature", e, SgdResponseStatus.SIGNATURE_NOT_VALID);
        }
    }

    public EncryptedMessageContainer keyDerivation(SgdRequest keyDerivationRequest) {
        performSignatureCheckAndThrowExceptionInCaseOfFailure(keyDerivationRequest);
        performGetPublicKeyVerificationAndThrowExceptionWhenFail(keyDerivationRequest, this.possibleCertificateSigners, possibleOcspSignerCertificates);
        checkIfOwnPublicKeyEciesIsInRequest(keyDerivationRequest);

        final KeyDerivationRequestSgdMessage cleartextRequestMessage
                = (KeyDerivationRequestSgdMessage) keyDerivationRequest.getEncryptedMessage().decryptToSgdMessage(currentHsmEciesKeyPair, previousHsmEciesKeyPair);

        performAtVerification(keyDerivationRequest, cleartextRequestMessage.getAtToken());

        final X509Certificate requestCertificate = keyDerivationRequest.getCertificateAsX509();
        String kvnr = SgdServerUtils.extractKVNRfromCertificate(requestCertificate).orElse("");
        String rnd = generate256BitRandomString();
        String telematikId = SgdCryptoUtilities.getInstance().extractTelematikIdFromCertificate(requestCertificate).orElse("");

        final SgdVector sgdVector = cleartextRequestMessage.getSgdVector();
        String associatedData = sgdVector.getAssociatedDataFunction()
                .getAssociatedData(sgdVector, rnd, this, kvnr, telematikId);
        byte[] currentDerivationKey = sgdVector.getDerivationKeyFunction()
                .getDerivationKey(sgdVector, this);

        byte[] aesKey = sgdCryptoUtilities.performHkdf(currentDerivationKey, associatedData.getBytes(), 256);

        String message = cleartextRequestMessage.getAtToken() + " "
                + cleartextRequestMessage.getRequestId()
                + " OK-KeyDerivation "
                + sgdCryptoUtilities.convertToLowerCaseHexWith64ByteLength(aesKey) + " "
                + associatedData;
        return buildEncryptedMessageContainerWithPayloadForRequest(message, keyDerivationRequest,
                Pair.of("kvnr", kvnr),
                Pair.of("rnd", rnd),
                Pair.of("associatedData", associatedData),
                Pair.of("key", sgdCryptoUtilities.convertByteArrayToHexString(aesKey)),
                Pair.of("derivationKeyName", currentDerivationKeyName),
                Pair.of("derivationKeyValue", sgdCryptoUtilities.convertByteArrayToHexString(currentDerivationKey))
        );
    }

    public SgdResponse receive(SgdRequest request) {
        if (request.getCommand() == SgdCommand.GET_PUBLIC_KEY) {
            return SgdResponse.create(getPublicKey(request));
        } else if (request.getCommand() == SgdCommand.GET_AUTHENTICATION_TOKEN) {
            return SgdResponse.create(getAuthenticationToken(request));
        } else if (request.getCommand() == SgdCommand.KEY_DERIVATION) {
            return SgdResponse.create(keyDerivation(request));
        } else {
            throw new IllegalArgumentException("Could not determine operation to be performed for command '" + request.getCommand() + "'!");
        }
    }

    public AesGcmEncryptionData offlineKeyDerivationForKvnr(String kvnr) {
        return offlineKeyDerivationForKvnr(SgdVector.createForR1RepeatedDerivation(kvnr, currentDerivationKeyName));
    }

    public AesGcmEncryptionData offlineKeyDerivationForKvnr(SgdVector sgdVector) {
        byte[] aesKey = sgdCryptoUtilities.performHkdf(getCurrentDerivationKeyValue(), sgdVector.toString().getBytes(), 256);

        return AesGcmEncryptionData.AesGcmEncryptionDataBuilder.anAesGcmEncryptionData()
                .withAesKey(aesKey)
                .withAssociatedData(sgdVector.toString())
                .build();
    }

    public void performEciesKeyChange() {
        this.previousHsmEciesKeyPair = currentHsmEciesKeyPair;
        this.currentHsmEciesKeyPair = sgdCryptoUtilities.generateEccKeyPair();

        this.previousAuthenticationTokenDerivationKey = currentAuthenticationTokenDerivationKey;
        this.currentAuthenticationTokenDerivationKey = generateAtDerivationKey();
    }

    @Override
    public String getCurrentDerivationKeyName() {
        return currentDerivationKeyName;
    }

    @Override
    public byte[] getCurrentDerivationKeyValue() {
        byte[] keyValue = this.availableDerivationKeys.get(currentDerivationKeyName);
        if (keyValue == null) {
            throw new RuntimeException("Unable to find current derivation Key!");
        }
        return keyValue;
    }

    @Override
    public Optional<byte[]> getDerivationKeyByName(String keyName) {
        return Optional.ofNullable(availableDerivationKeys.get(keyName));
    }
}
