/*
 Copyright (c) 2020 gematik GmbH

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */
package de.gematik.ti.sgd.domain.json;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.annotation.JsonProperty;
import de.gematik.ti.sgd.domain.sgdCommunication.EncryptedMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.PublicKeyEcies;
import lombok.AllArgsConstructor;
import lombok.Builder;
import lombok.Data;
import lombok.NoArgsConstructor;

import java.util.HashMap;
import java.util.Map;

/**
 * Aggregierte Klasse von EncryptedMessageContainer und GetPublicKeyResponse.
 * Viele JSON-Serialisierer haben Probleme mit Polymorphie, Mehrfachvererbung gibt es in Java nicht.
 * Daher enthält diese Klasse manuell alle Felder der beiden tatsächlichen Message-Klassen und mappt dann.
 */
@Data
@AllArgsConstructor
@NoArgsConstructor
@Builder
@JsonInclude(JsonInclude.Include.NON_NULL)
public class SgdResponse {
    // GetPublicKey
    private PublicKeyEcies PublicKeyECIES;
    private String Signature;
    private String Certificate;
    // EncryptedMessageContainer
    private SgdResponseStatus status;
    private String EncryptedMessage;
    private String debug_encryptedMessageCleartext;
    private String debug_clientEphemeralKey_Public;
    private String debug_senderKey_Public;
    private Map<String, String> debug_data;
    // Error Messages
    private String errorMessage;
    private String errorDetails;

    @JsonProperty("PublicKeyECIES")
    public PublicKeyEcies getPublicKeyECIES() {
        return PublicKeyECIES;
    }

    @JsonProperty("Signature")
    public String getSignature() {
        return Signature;
    }

    @JsonProperty("Certificate")
    public String getCertificate() {
        return Certificate;
    }

    @JsonProperty("Status")
    public SgdResponseStatus getStatus() {
        return status;
    }

    @JsonProperty("EncryptedMessage")
    public String getEncryptedMessage() {
        return EncryptedMessage;
    }

    public static SgdResponse create(GetPublicKeyResponse publicKey) {
        return SgdResponse.builder()
                .PublicKeyECIES(publicKey.getPublicKeyEcies())
                .Signature(publicKey.getSignature())
                .Certificate(publicKey.getCertificate())
                .build();
    }

    public static SgdResponse create(EncryptedMessageContainer encryptedMessageContainer) {
        return SgdResponse.builder()
                .debug_encryptedMessageCleartext(encryptedMessageContainer.getDebug_encryptedMessageCleartext())
                .debug_clientEphemeralKey_Public(encryptedMessageContainer.getDebug_clientEphemeralKey_Public())
                .debug_senderKey_Public(encryptedMessageContainer.getDebug_senderKey_Public())
                .debug_data(encryptedMessageContainer.getDebug_data())
                .status(encryptedMessageContainer.getStatus())
                .EncryptedMessage(encryptedMessageContainer.getEncryptedMessage().getContent())
                .build();
    }

    public EncryptedMessageContainer toEncryptedMessageContainer() {
        return EncryptedMessageContainer.EncryptedMessageContainerBuilder
                .anEncryptedMessageContainer()
                .withEncryptedMessage(new EncryptedMessage(EncryptedMessage))
                .withStatus(status)
                .withDebug_clientEphemeralKey_Public(debug_clientEphemeralKey_Public)
                .withDebug_data(debug_data)
                .withDebug_encryptedMessageCleartext(debug_encryptedMessageCleartext)
                .withDebug_senderKey_Public(debug_senderKey_Public)
                .build();
    }

    public GetPublicKeyResponse toGetPublicKeyResponse() {
        return GetPublicKeyResponse.builder()
                .certificate(Certificate)
                .publicKeyEcies(PublicKeyECIES)
                .signature(Signature)
                .build();
    }

    public void addDebugData(String key, String value) {
        if (getDebug_data() == null) {
            setDebug_data(new HashMap<>());
        }
        debug_data.put(key, value);
    }
}
