/*
 Copyright (c) 2020 gematik GmbH

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 */
package de.gematik.ti.sgd.domain.sgdCommunication;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.JsonSerializer;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import lombok.Data;
import org.apache.commons.codec.digest.DigestUtils;

import java.io.IOException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.interfaces.RSAPrivateKey;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@Data
@JsonSerialize(using = PublicKeyEcies.PublicKeyEciesSerializer.class)
public class PublicKeyEcies {
    private final String message;
    private final String[] parts;

    public PublicKeyEcies(String message) {
        this.message = message;
        parts = message.split(" ");
    }

    public PublicKeyEcies(PublicKey publicKey) {
        message = convertPublicKeyToGemSpecString(publicKey);
        parts = message.split(" ");
    }

    public PublicKeyEcies(PublicKey publicKey, PublicKey sgd1PublicKey, PublicKey sgd2PublicKey) {
        String curveName = "brainpoolP256r1";
        message = curveName
                + " " + SgdCryptoUtilities.getInstance().extractPublicPointCoordinatesFromPublicKey(publicKey)
                + " " + DigestUtils.sha256Hex(convertPublicKeyToGemSpecString(sgd1PublicKey))
                + " " + DigestUtils.sha256Hex(convertPublicKeyToGemSpecString(sgd2PublicKey));
        parts = message.split(" ");
    }

    private String convertPublicKeyToGemSpecString(PublicKey publicKey) {
        String message;
        String curveName = "brainpoolP256r1";
        message = curveName + " " + SgdCryptoUtilities.getInstance().extractPublicPointCoordinatesFromPublicKey(publicKey);
        return message;
    }

    public byte[] getBytes() {
        return message.getBytes();
    }

    public PublicKey getPublicKey() {
        return SgdCryptoUtilities.getInstance().extractPublicKeyFromPublicPointCoordinates(
                parts[1].replace("0x", ""),
                parts[2].replace("0x", ""),
                parts[0]
        );
    }

    public String getSignatureAsBase64EncodedString(PrivateKey signerPrivateKey) {
        byte[] signatureValue = null;
        if (signerPrivateKey instanceof RSAPrivateKey) {
            signatureValue = SgdCryptoUtilities.getInstance().signRsa(getBytes(), signerPrivateKey);
        } else {
            signatureValue = SgdCryptoUtilities.getInstance().signEcdsa(getBytes(), signerPrivateKey);
        }
        return SgdCryptoUtilities.getInstance().convertBytesToB64String(signatureValue);
    }

    public String getSgd1PublicKeyHash() {
        return parts[3];
    }

    public String getSgd2PublicKeyHash() {
        return parts[4];
    }

    public String getEciesPublicKeyAsString() {
        return Stream.of(message.split(" "))
                .limit(3)
                .collect(Collectors.joining(" "));
    }

    public static class PublicKeyEciesSerializer extends JsonSerializer<PublicKeyEcies> {
        @Override
        public void serialize(PublicKeyEcies value, JsonGenerator jsonGenerator, SerializerProvider serializerProvider) throws IOException {
            jsonGenerator.writeString(value.getMessage());
        }
    }
}
