/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.server;

import com.jayway.jsonpath.DocumentContext;
import com.jayway.jsonpath.JsonPath;
import de.gematik.ti.sgd.components.SgdClientKeyDerivationMachine;
import de.gematik.ti.sgd.components.SgdClientUtils;
import de.gematik.ti.sgd.domain.SgdInstance;
import org.junit.Before;
import org.junit.Test;
import org.springframework.boot.test.web.client.TestRestTemplate;
import org.springframework.http.ResponseEntity;
import org.springframework.test.util.ReflectionTestUtils;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.DefaultUriBuilderFactory;

import java.security.KeyPair;
import java.security.PublicKey;
import java.util.List;

import static org.assertj.core.api.Assertions.assertThat;

public class BasicCommunicationTest extends TestWithConfiguredSgdClient {
    @Test
    public void getPublicKeyFromServer_checkJsonStructure() {
        final ResponseEntity<String> responseEntity = testRestTemplate.postForEntity("/sgd1",
                SgdClientUtils.generateGetPublicKeyRequest(clientCertificate),
                String.class);

        assertThat(responseEntity.getStatusCode().is2xxSuccessful())
                .isTrue();

        final DocumentContext result = JsonPath.parse(responseEntity.getBody());

        assertThat(result.read("@.PublicKeyECIES", String.class))
                .startsWith("brainpoolP256r1 0x");
        assertThat(result.read("@.Signature", String.class))
                .isNotEmpty();
        assertThat(result.read("@.Certificate", String.class))
                .startsWith("MIIC");
    }

    @Test
    public void getAuthenticationToken_checkJsonStructure() {
        SgdClientKeyDerivationMachine stateMachine = new SgdClientKeyDerivationMachine(clientCertificate,
                (KeyPair) ReflectionTestUtils.getField(sgdClient, "clientKeyPair"),
                (List<PublicKey>) ReflectionTestUtils.getField(sgdClient, "possibleSigners"));

        stateMachine.receivePublicKeyResponse_sgd1(
                ReflectionTestUtils.invokeMethod(sgdClient, "performPublicKeyExchange",
                        stateMachine.getPublicKeyRequest(), SgdInstance.SGD_1));
        stateMachine.receivePublicKeyResponse_sgd2(
                ReflectionTestUtils.invokeMethod(sgdClient, "performPublicKeyExchange",
                        stateMachine.getPublicKeyRequest(), SgdInstance.SGD_2));

        final ResponseEntity<String> responseEntity = testRestTemplate.postForEntity("/sgd1",
                stateMachine.getAuthenticationTokenRequest_sgd1(),
                String.class);

        final DocumentContext result = JsonPath.parse(responseEntity.getBody());

        assertThat(result.read("@.EncryptedMessage", String.class))
                .startsWith("brainpoolP256r1 ");
        assertThat(result.read("@.Status", String.class))
                .isEqualTo("OK");
    }

    @Test
    public void performKeyDerivation() {
        assertThat(sgdClient.getSgd1AesGcmKey())
                .isNull();

        sgdClient.performInitialSgdKeyDerivation();

        assertThat(sgdClient.getSgd1AesGcmKey())
                .isNotNull();
    }
}